var Collection, IndexedDb, async, compileSort, createUid, processFind, _;

_ = require('lodash');

async = require('async');

createUid = require('./utils').createUid;

processFind = require('./utils').processFind;

compileSort = require('./selector').compileSort;

module.exports = IndexedDb = (function() {
  function IndexedDb(options, success, error) {
    var createTables;
    this.collections = {};
    this.db = window.openDatabase('minimongo_' + options.namespace, '1.0', 'Minimongo:' + options.namespace, 50 * 1024 * 1024);
    if (!this.db) {
      return error("Failed to create database");
    }
    createTables = (function(_this) {
      return function(tx) {
        return tx.executeSql('CREATE TABLE IF NOT EXISTS docs (\n  col TEXT NOT NULL,\n  id TEXT NOT NULL,\n  state TEXT NOT NULL,\n  doc TEXT, \n  PRIMARY KEY (col, id));');
      };
    })(this);
    this.db.transaction(createTables, error, (function(_this) {
      return function() {
        if (success) {
          return success(_this);
        }
      };
    })(this));
  }

  IndexedDb.prototype.addCollection = function(name, success, error) {
    var collection;
    collection = new Collection(name, this.db);
    this[name] = collection;
    this.collections[name] = collection;
    if (success) {
      return success();
    }
  };

  IndexedDb.prototype.removeCollection = function(name, success, error) {
    delete this[name];
    delete this.collections[name];
    return this.db.transaction((function(_this) {
      return function(tx) {
        return tx.executeSql("DELETE FROM docs WHERE col = ?", [name], success);
      };
    })(this), error);
  };

  return IndexedDb;

})();

Collection = (function() {
  function Collection(name, db) {
    this.name = name;
    this.db = db;
  }

  Collection.prototype.find = function(selector, options) {
    return {
      fetch: (function(_this) {
        return function(success, error) {
          return _this._findFetch(selector, options, success, error);
        };
      })(this)
    };
  };

  Collection.prototype.findOne = function(selector, options, success, error) {
    var _ref;
    if (_.isFunction(options)) {
      _ref = [{}, options, success], options = _ref[0], success = _ref[1], error = _ref[2];
    }
    return this.find(selector, options).fetch(function(results) {
      if (success != null) {
        return success(results.length > 0 ? results[0] : null);
      }
    }, error);
  };

  Collection.prototype._findFetch = function(selector, options, success, error) {
    return this.db.transaction((function(_this) {
      return function(tx) {
        return tx.executeSql("SELECT * FROM docs WHERE col = ?", [_this.name], function(tx, results) {
          var docs, i, row, _i, _ref;
          docs = [];
          for (i = _i = 0, _ref = results.rows.length; 0 <= _ref ? _i < _ref : _i > _ref; i = 0 <= _ref ? ++_i : --_i) {
            row = results.rows.item(i);
            if (row.state !== "removed") {
              docs.push(JSON.parse(row.doc));
            }
          }
          if (success != null) {
            return success(processFind(docs, selector, options));
          }
        });
      };
    })(this), error);
  };

  Collection.prototype.upsert = function(doc, success, error) {
    if (!doc._id) {
      doc._id = createUid();
    }
    return this.db.transaction((function(_this) {
      return function(tx) {
        return tx.executeSql("INSERT OR REPLACE INTO docs (col, id, state, doc) VALUES (?, ?, ?, ?)", [_this.name, doc._id, "upserted", JSON.stringify(doc)], function() {
          if (success) {
            return success(doc);
          }
        });
      };
    })(this), error);
  };

  Collection.prototype.remove = function(id, success, error) {
    return this.db.transaction((function(_this) {
      return function(tx) {
        return tx.executeSql("SELECT * FROM docs WHERE col = ? AND id = ?", [_this.name, id], function(tx, results) {
          if (results.rows.length > 0) {
            return tx.executeSql('UPDATE docs SET state="removed" WHERE col = ? AND id = ?', [_this.name, id], function() {
              if (success) {
                return success(id);
              }
            });
          } else {
            return tx.executeSql("INSERT INTO docs (col, id, state, doc) VALUES (?, ?, ?, ?)", [
              _this.name, id, "removed", JSON.stringify({
                _id: id
              })
            ], function() {
              if (success) {
                return success(id);
              }
            });
          }
        });
      };
    })(this), error);
  };

  Collection.prototype.cache = function(docs, selector, options, success, error) {
    return async.each(docs, (function(_this) {
      return function(doc, callback) {
        return _this.db.transaction(function(tx) {
          return tx.executeSql("SELECT * FROM docs WHERE col = ? AND id = ?", [_this.name, doc._id], function(tx, results) {
            if (results.rows.length === 0 || results.rows.item(0).state === "cached") {
              return tx.executeSql("INSERT OR REPLACE INTO docs (col, id, state, doc) VALUES (?, ?, ?, ?)", [_this.name, doc._id, "cached", JSON.stringify(doc)], function() {
                return callback();
              });
            } else {
              return callback();
            }
          });
        }, callback);
      };
    })(this), (function(_this) {
      return function(err) {
        var docsMap, sort;
        if (err) {
          if (error) {
            error(err);
          }
          return;
        }
        docsMap = _.object(_.pluck(docs, "_id"), docs);
        if (options.sort) {
          sort = compileSort(options.sort);
        }
        return _this.find(selector, options).fetch(function(results) {
          return async.each(results, function(result, callback) {
            return _this.db.transaction(function(tx) {
              return tx.executeSql("SELECT * FROM docs WHERE col = ? AND id = ?", [_this.name, result._id], function(tx, rows) {
                if (!docsMap[result._id] && rows.rows.length > 0 && rows.rows.item(0).state === "cached") {
                  if (options.sort && options.limit && docs.length === options.limit) {
                    if (sort(result, _.last(docs)) >= 0) {
                      return callback();
                    }
                  }
                  return tx.executeSql("DELETE FROM docs WHERE col = ? AND id = ?", [_this.name, result._id], function() {
                    return callback();
                  });
                } else {
                  return callback();
                }
              });
            }, callback);
          }, function(err) {
            if (err != null) {
              if (error != null) {
                error(err);
              }
              return;
            }
            if (success != null) {
              return success();
            }
          });
        }, error);
      };
    })(this));
  };

  Collection.prototype.pendingUpserts = function(success, error) {
    return this.db.transaction((function(_this) {
      return function(tx) {
        return tx.executeSql("SELECT * FROM docs WHERE col = ? AND state = ?", [_this.name, "upserted"], function(tx, results) {
          var docs, i, row, _i, _ref;
          docs = [];
          for (i = _i = 0, _ref = results.rows.length; 0 <= _ref ? _i < _ref : _i > _ref; i = 0 <= _ref ? ++_i : --_i) {
            row = results.rows.item(i);
            docs.push(JSON.parse(row.doc));
          }
          if (success != null) {
            return success(docs);
          }
        });
      };
    })(this), error);
  };

  Collection.prototype.pendingRemoves = function(success, error) {
    return this.db.transaction((function(_this) {
      return function(tx) {
        return tx.executeSql("SELECT * FROM docs WHERE col = ? AND state = ?", [_this.name, "removed"], function(tx, results) {
          var docs, i, row, _i, _ref;
          docs = [];
          for (i = _i = 0, _ref = results.rows.length; 0 <= _ref ? _i < _ref : _i > _ref; i = 0 <= _ref ? ++_i : --_i) {
            row = results.rows.item(i);
            docs.push(JSON.parse(row.doc)._id);
          }
          if (success != null) {
            return success(docs);
          }
        });
      };
    })(this), error);
  };

  Collection.prototype.resolveUpsert = function(doc, success, error) {
    return this.db.transaction((function(_this) {
      return function(tx) {
        return tx.executeSql("SELECT * FROM docs WHERE col = ? AND id = ?", [_this.name, doc._id], function(tx, results) {
          if (results.rows.length > 0) {
            if (results.rows.item(0).state === "upserted" && _.isEqual(JSON.parse(results.rows.item(0).doc), doc)) {
              return tx.executeSql('UPDATE docs SET state="cached" WHERE col = ? AND id = ?', [_this.name, doc._id], function() {
                if (success) {
                  return success(doc);
                }
              });
            } else {
              if (success) {
                return success(doc);
              }
            }
          } else {
            return error(new Error("Upsert not found"));
          }
        });
      };
    })(this), error);
  };

  Collection.prototype.resolveRemove = function(id, success, error) {
    return this.db.transaction((function(_this) {
      return function(tx) {
        return tx.executeSql('DELETE FROM docs WHERE state="removed" AND col = ? AND id = ?', [_this.name, id], function() {
          if (success) {
            return success(id);
          }
        });
      };
    })(this), error);
  };

  Collection.prototype.seed = function(doc, success, error) {
    return this.db.transaction((function(_this) {
      return function(tx) {
        return tx.executeSql("SELECT * FROM docs WHERE col = ? AND id = ?", [_this.name, doc._id], function(tx, results) {
          if (results.rows.length === 0) {
            return tx.executeSql("INSERT INTO docs (col, id, state, doc) VALUES (?, ?, ?, ?)", [_this.name, doc._id, "cached", JSON.stringify(doc)], function() {
              if (success) {
                return success(doc);
              }
            });
          } else {
            if (success) {
              return success(doc);
            }
          }
        });
      };
    })(this), error);
  };

  Collection.prototype.cacheOne = function(doc, success, error) {
    return this.db.transaction((function(_this) {
      return function(tx) {
        return tx.executeSql("SELECT * FROM docs WHERE col = ? AND id = ?", [_this.name, doc._id], function(tx, results) {
          if (results.rows.length === 0 || results.rows.item(0).state === "cached") {
            return tx.executeSql("INSERT OR REPLACE INTO docs (col, id, state, doc) VALUES (?, ?, ?, ?)", [_this.name, doc._id, "cached", JSON.stringify(doc)], function() {
              if (success) {
                return success(doc);
              }
            });
          } else {
            if (success) {
              return success(doc);
            }
          }
        });
      };
    })(this), error);
  };

  return Collection;

})();
